from .elementdb import ElementDB

class Element(object):
    """ Element from periodic table

    """
    def __init__(self, identifier, mass=None):
        """ Initializes element from identifier and mass

        :param str or int identifier: Symbol, Name, or Atomic Number of element
        :param float mass: Mass [amu] of element
        """
        element = ElementDB.lookup(identifier)

        self._symbol = element['symbol']
        self._name = element['name']
        self._atomic_number = element['z']

        if mass:
            self._mass = mass
        else:
            self._mass = element['mass']
            
    def __eq__(self, element):
        if (self.symbol == element.symbol and
            self.name == element.name and
            self.atomic_number == element.atomic_number and
            self.mass == element.mass):
            return True
        return False

    def __repr__(self):
        return "<Element symbol:{} name:{} mass:{:2.2f}>".format(
            self.symbol, self.name, self.mass)

    def __hash__(self):
        return sum(hash(item) for item in [
            self._mass, self._symbol, self._name, self.atomic_number
        ])

    @property
    def symbol(self):
        return self._symbol

    @property
    def name(self):
        return self._name

    @property
    def atomic_number(self):
        return self._atomic_number

    @property
    def mass(self):
        return self._mass




from .material import Material
from .utils import check_input, is_positive

class Layer(Material):
    """ Represents a layer in target

    """
    def __init__(self, elements, density, width, phase=0, name=None):
        """ Creation of Layer from elements, density, width, phase, and name

        :params float width: width [Ang] of layer
        :params str name: Name of the Layer (defaults to chemical_formula)
        
        See Material for documentation on elements, density, and phase parameters
        """
        self.width = width
        self.name = name
        super(Layer, self).__init__(elements, density, phase)

    @classmethod
    def from_formula(cls, chemical_formula, density, width, phase=0, name=None):
        """ Creation Layer from chemical formula string, density, width, phase, and name

        :params float width: width [Ang] of layer
        :params str name: Name of the Layer (defaults to chemical_formula)

        See Material for documentation on chemical_formula, density, and phase parameters
        """
        elements = cls._formula_to_elements(chemical_formula)
        return Layer(elements, density, width, phase, name)

    @property
    def width(self):
        return self._width

    @width.setter
    def width(self, value):
        self._width = check_input(float, is_positive, value)

    @property
    def name(self):
        if self._name:
            return self._name
        return self.chemical_formula

    @name.setter
    def name(self, value):
        self._name = str(value)

    def __repr__(self):
        return "<Layer material:{} width:{}>".format(self.chemical_formula, self.width)



""" Module for automating srim calculations

"""
import os
import subprocess
import shutil

from .core.utils import (
    check_input,
    is_zero, is_zero_or_one, is_zero_to_two, is_zero_to_five, is_one_to_seven, is_one_to_eight,
    is_srim_degrees,
    is_positive,
    is_quoteless
)

from .output import Results
from .input import AutoTRIM, TRIMInput, SRInput


SRIM_DIRECTORY = os.path.join(os.sep, 'tmp', 'srim')


class SRIMSettings(object):
    """ SRIM Settings

    TODO: Readonly becuase I have not constructed getter and setters
    """
    def __init__(self, **args):
        self._settings = {
            'description': check_input(str, is_quoteless, args.get('description', 'srim-python run')),
            'reminders': check_input(int, is_zero_or_one, args.get('reminders', 0)),
            'autosave': check_input(int, is_zero_or_one, args.get('autosave', 0)),
            'plot_mode': check_input(int, is_zero_to_five, args.get('plot_mode', 5)),
            'plot_xmin': check_input(float, is_positive, args.get('plot_xmin', 0.0)),
            'plot_xmax': check_input(float, is_positive, args.get('plot_xmax', 0.0)),
            'ranges': check_input(int, is_zero_or_one, args.get('ranges', 0)),
            'backscattered': check_input(int, is_zero_or_one, args.get('backscattered', 0)),
            'transmit': check_input(int, is_zero_or_one, args.get('transmit', 0)),
            'sputtered': check_input(int, is_zero_or_one, args.get('ranges', 0)),
            'collisions': check_input(int, is_zero_to_two, args.get('collisions', 0)),
            'exyz': check_input(int, is_zero_or_one, args.get('exyz', 0)),
            'angle_ions': check_input(float, is_srim_degrees, args.get('angle_ions', 0.0)),
            'bragg_correction': float(args.get('bragg_correction', 1.0)), # TODO: Not sure what correct values are
            'random_seed': check_input(int, is_positive, args.get('random_seed', 0)),
            'version': check_input(int, is_zero_or_one, args.get('version', 0)),
        }

        if self.plot_xmin > self.plot_xmax:
            raise ValueError('xmin must be <= xmax')

    def __getattr__(self, attr):
        return self._settings[attr]


class SRIM(object):
    """ Automate SRIM Calculations

    """
    def __init__(self, target, ion, calculation=1, number_ions=1000, **args):
        """ Initialize srim object with settings

        TODO: fill out doc strings
        """
        self.settings = SRIMSettings(**args)
        self.calculation = check_input(int, is_one_to_seven, calculation)
        self.number_ions = check_input(int, is_positive, number_ions)
        self.target = target
        self.ion = ion


    def _write_input_files(self):
        """ Write necissary TRIM input files for calculation """
        AutoTRIM().write()
        TRIMInput(self).write()

    @staticmethod
    def copy_output_files(src_directory, dest_directory, check_srim_output=True):
        known_files = {
            'TRIM.IN', 'PHONON.txt', 'E2RECOIL.txt', 'IONIZ.txt',
            'LATERAL.txt', 'NOVAC.txt', 'RANGE.txt', 'VACANCY.txt',
            'COLLISON.txt', 'BACKSCAT.txt', 'SPUTTER.txt',
            'RANGE_3D.txt', 'TRANSMIT.txt', 'TRIMOUT.txt'
        }

        if not os.path.isdir(src_directory):
            raise ValueError('src_directory must be path')

        if not os.path.isdir(dest_directory):
            raise ValueError('dest_directory must be path')

        for known_file in known_files:
            if os.path.isfile(os.path.join(
                    src_directory, known_file)):
                shutil.copy(os.path.join(
                    src_directory, known_file), dest_directory)
            elif os.path.isfile(os.path.join(src_directory, 'SRIM Outputs', known_file)) and check_srim_output:
                shutil.copy(os.path.join(
                    src_directory, 'SRIM Outputs', known_file), dest_directory)

    def run(self, srim_directory=SRIM_DIRECTORY):
        current_directory = os.getcwd()
        os.chdir(srim_directory)
        self._write_input_files()
        # Make sure compatible with Windows, OSX, and Linux
        subprocess.check_call(['wine', str(os.path.join('.', 'TRIM.exe'))])
        os.chdir(current_directory)

        return Results(srim_directory)


class SRSettings(object):
    """ SR Settings """
    def __init__(self, **args):
        self._settings = {
            'energy_min': check_input(float, is_positive, args.get('energy_min', 1.0E3)),
            'output_type': check_input(int, is_one_to_eight, args.get('output_type', 1)),
            'output_filename': args.get('output_filename', 'SR_OUTPUT.txt'),
            'correction': check_input(float, is_positive, args.get('correction', 1.0))
        }

    def __getattr__(self, attr):
        return self._settings[attr]


class SR(object):
    """ Automate SR Calculations """
    def __init__(self, layer, ion, **args):
        self.settings = SRSettings(**args)
        self.layer = layer
        self.ion = ion

    def  _write_input_file(self):
        """ Write necissary SR input file for calculation """
        SRInput(self).write()

    def run(self, srim_directory=SRIM_DIRECTORY):
        current_directory = os.getcwd()
        os.chdir(os.path.join(srim_directory, 'SR Module'))
        self._write_input_file()
        # Make sure compatible with Windows, OSX, and Linux
        subprocess.check_call([str(os.path.join('.', 'SRModule.exe'))])
        os.chdir(current_directory)



from math import sqrt

from . import units
from .element import Element

class Ion(Element):
    """ Ion traveling through medium

    """
    def __init__(self, identifier, energy, mass=None):
        """Initialize Ion

        :param int or str symbol: Atomic symbol of element
        :param float energy: Energy [eV] of ion
        :param float mass: Mass [amu] of ion
        """
        if energy <= 0.0:
            raise ValueError('energy {} cannot be 0.0 or less'.format(energy))

        self._energy = energy
        super(Ion, self).__init__(identifier, mass)

    def __repr__(self):
        return "<Ion element:{} mass:{} energy:{} keV>".format(
            self.name, self.mass, self.energy)

    @property
    def energy(self):
        return self._energy

    @property
    def velocity(self):
        """ Velocity of Ion [m/s] """
        return sqrt(2 * (self.energy * units.eV) / (self.mass * units.amu))




""" Read output files of SRIM simulation

TODO: Read header information
"""
import os
import re
from io import BytesIO

import numpy as np

from .core.ion import Ion

# Valid double_regex [4, 4.0, 4.0e100
double_regex = r'[-+]?\d+\.?\d*(?:[eE][-+]?\d+)?'
symbol_regex = r'[A-Z][a-z]?'
int_regex = '[+-]?\d+'


class SRIMOutputParseError(Exception):
    pass


class SRIM_Output(object):
    def _read_name(self, output):
        raise NotImplementedError()

    def _read_ion(self, output):
        ion_regex = 'Ion\s+=\s+({})\s+Energy\s+=\s+({})\s+keV'.format(
            symbol_regex, double_regex)
        match = re.search(ion_regex.encode('utf-8'), output)
        if match:
            symbol = str(match.group(1).decode('utf-8'))
            energy = float(match.group(2)) #keV
            return Ion(symbol, 1000.0 * energy)
        raise SRIMOutputParseError("unable to extract ion from file")

    def _read_target(self, output):
        match_target = re.search(b'(?<=====\r\n)Layer\s+\d+\s+:.*?(?=====)', output, re.DOTALL)
        if match_target:
            print(match_target.group(0))
            layer_regex = (
                'Layer\s+(?P<i>\d+)\s+:\s+(.+)\r\n'
                'Layer Width\s+=\s+({0})\s+A\s+;\r\n'
                '\s+Layer #\s+(?P=i)- Density = ({0}) atoms/cm3 = ({0}) g/cm3\r\n'
                '((?:\s+Layer #\s+(?P=i)-\s+{1}\s+=\s+{0}\s+Atomic Percent = {0}\s+Mass Percent\r\n)+)'
            ).format(double_regex, symbol_regex)
            layers = re.findall(layer_regex.encode('utf-8'), match_target.group(0))
            if layers:
                element_regex = (
                    '\s+Layer #\s+(\d+)-\s+({1})\s+=\s+({0})\s+Atomic Percent = ({0})\s+Mass Percent\r\n'
                ).format(double_regex, symbol_regex)
                element_regex = element_regex.encode()

                layers_elements = []
                for layer in layers:
                    # We know that elements will match
                    layers_elements.append(re.findall(element_regex, layer[5]))

                raise NotImpementedError()

                import pytest
                pytest.set_trace()

        raise SRIMOutputParseError("unable to extract total target from file")

    def _read_num_ions(self, output):
        match = re.search(b'Total Ions calculated\s+=(\d+.\d+)', output)
        if match:
            # Cast string -> float -> round down to nearest int
            return int(float(match.group(1)))
        raise SRIMOutputParseError("unable to extract total ions from file")

    def _read_table(self, output):
        match = re.search((
            b'=+(.*)'
            b'-+(?:\s+-+)+'
        ), output, re.DOTALL)
        # Read Data from table

        if match:
            # Headers TODO: name the columns in table
            header = None

            # Data
            data = np.genfromtxt(BytesIO(output[match.end():]), max_rows=100)
            return data
        raise SRIMOutputParseError("unable to extract table from file")


class Results(object):
    """ Gathers all results from folder """
    def __init__(self, directory):
        """ Retrives all the calculation files in a given folder

        Args:
          directory (str): directory that srim output files are located
        """
        self.ioniz = Ioniz(directory)
        self.vacancy = Vacancy(directory)

        try:
            self.novac = NoVacancy(directory)
        except ValueError:
            self.novac = None

        self.etorecoils = EnergyToRecoils(directory)
        self.phonons = Phonons(directory)
        self.range = Range(directory)


class Ioniz(SRIM_Output):
    def __init__(self, directory, filename='IONIZ.txt'):
        with open(os.path.join(directory, filename), 'rb') as f:
            output = f.read()
            ion = self._read_ion(output)
            num_ions = self._read_num_ions(output)
            data = self._read_table(output)

        self._ion = ion
        self._num_ions = num_ions
        self._depth = data[:, 0]
        self._ions = data[:, 1]
        self._recoils = data[:, 2]

    @property
    def ion(self):
        """ Ion used in SRIM calculation

        **mass** could be wrong
        """
        return self._ion

    @property
    def num_ions(self):
        """ Number of Ions in SRIM simulation """
        return self._num_ions

    @property
    def depth(self):
        """ Depth [Ang] of bins in SRIM Calculation """
        return self._depth

    @property
    def ions(self):
        """Ionization energy [eV/(Angstrom Ion)] lost to electronic stopping
        in incident ions

        """
        return self._ions

    @property
    def recoils(self):
        """Ionization energy [eV/(Angstrom Ion)] lost to electronic stopping
        in recoil ions
        """
        return self._recoils


class Vacancy(SRIM_Output):
    """ Table of the final distribution of vacancies """
    def __init__(self, directory, filename='VACANCY.txt'):
        with open(os.path.join(directory, filename), 'rb') as f:
            output = f.read()
            ion = self._read_ion(output)
            num_ions = self._read_num_ions(output)
            data = self._read_table(output)

        self._ion = ion
        self._num_ions = num_ions
        self._depth = data[:, 0]
        self._ion_knock_ons = data[:, 1]
        self._vacancies = data[:, 2:]

    @property
    def ion(self):
        """ Ion used in SRIM calculation

        **mass** could be wrong
        """
        return self._ion

    @property
    def num_ions(self):
        """ Number of Ions in SRIM simulation """
        return self._num_ions

    @property
    def depth(self):
        """ Depth [Ang] of bins in SRIM Calculation """
        return self._depth

    @property
    def knock_ons(self):
        """ Vacancies produced [Vacancies/(Angstrom-Ion) by ion] """
        return self._ion_knock_ons

    @property
    def vacancies(self):
        """ Vacancies [Vacancies/(Angstrom-Ion)] produced of element in layer

        TODO: improve interface
        """
        return self._vacancies


class NoVacancy(SRIM_Output):
    """ Table of Replacement Collisions """
    def __init__(self, directory, filename='NOVAC.txt'):
        with open(os.path.join(directory, filename), 'rb') as f:
            output = f.read()

            # Check if it is KP calculation
            if re.search(b'Recoil/Damage Calculations made with Kinchin-Pease Estimates',
                         output):
                raise ValueError('NOVAC has no data for KP calculations')

            ion = self._read_ion(output)
            num_ions = self._read_num_ions(output)
            data = self._read_table(output)

        self._ion = ion
        self._num_ions = num_ions
        self._depth = data[:, 0]
        self._number = data[:, 1]

    @property
    def ion(self):
        """ Ion used in SRIM calculation

        **mass** could be wrong
        """
        return self._ion

    @property
    def num_ions(self):
        """ Number of Ions in SRIM simulation """
        return self._num_ions

    @property
    def depth(self):
        """ Depth [Ang] of bins in SRIM Calculation """
        return self._depth

    @property
    def number(self):
        """ Replacement Collisions [Number/(Angstrom-Ion)]"""
        return self._number


class EnergyToRecoils(SRIM_Output):
    """ Energy transfered to atoms through binary collision """
    def __init__(self, directory, filename='E2RECOIL.txt'):
        with open(os.path.join(directory, filename), 'rb') as f:
            output = f.read()
            ion = self._read_ion(output)
            num_ions = self._read_num_ions(output)
            data = self._read_table(output)

        self._ion = ion
        self._num_ions = num_ions
        self._depth = data[:, 0]
        self._ions = data[:, 1]
        self._recoils = data[:, 2:]

    @property
    def ion(self):
        """ Ion used in SRIM calculation

        **mass** could be wrong
        """
        return self._ion

    @property
    def num_ions(self):
        """ Number of Ions in SRIM simulation """
        return self._num_ions

    @property
    def depth(self):
        """ Depth [Ang] of bins in SRIM Calculation """
        return self._depth

    @property
    def ions(self):
        """ Energy [eV/(Angstrom-Ion)] transfered to material through ion collisions """
        return self._ions

    @property
    def absorbed(self):
        """ Energy [eV/(Angstrom-Ion)] absorbed from collisions with Atom

        TODO: fix terminology
        """
        return self._recoils


class Phonons(SRIM_Output):
    """ Distribution of Phonons """
    def __init__(self, directory, filename='PHONON.txt'):
        with open(os.path.join(directory, filename), 'rb') as f:
            output = f.read()
            ion = self._read_ion(output)
            num_ions = self._read_num_ions(output)
            data = self._read_table(output)

        self._ion = ion
        self._num_ions = num_ions
        self._depth = data[:, 0]
        self._ions = data[:, 1]
        self._recoils = data[:, 2]

    @property
    def ion(self):
        """ Ion used in SRIM calculation

        **mass** could be wrong
        """
        return self._ion

    @property
    def num_ions(self):
        """ Number of Ions in SRIM simulation """
        return self._num_ions

    @property
    def depth(self):
        """ Depth [Ang] of bins in SRIM Calculation """
        return self._depth

    @property
    def ions(self):
        """ Number of phonons [Phonons/(Angstrom Ion)] created from ions collisions

        """
        return self._ions

    @property
    def recoils(self):
        """Number of phonons [Phonons/(Angstrom Ion)] created from recoils
        resulting from ion collisions

        """
        return self._recoils


class Range(SRIM_Output):
    """ Table of the final distribution of the ions, and any recoiling target atoms
    """
    def __init__(self, directory, filename='RANGE.txt'):
        with open(os.path.join(directory, filename), 'rb') as f:
            output = f.read()
            ion = self._read_ion(output)
            num_ions = self._read_num_ions(output)
            data = self._read_table(output)

        self._ion = ion
        self._num_ions = num_ions
        self._depth = data[:, 0]
        self._ions = data[:, 1]
        self._elements = data[:, 2:]

    @property
    def ion(self):
        """ Ion used in SRIM calculation

        **mass** could be wrong
        """
        return self._ion

    @property
    def num_ions(self):
        """ Number of Ions in SRIM simulation """
        return self._num_ions

    @property
    def depth(self):
        """ Depth [Ang] of bins in SRIM Calculation """
        return self._depth

    @property
    def ions(self):
        """ Ion final distribution [(Atoms/cm3)/(Atoms/cm2)] """
        return self._ions

    @property
    def elements(self):
        """ Per elements [(Atoms/cm3)/(Atoms/cm2)] distribution of each element

        """
        return self._elements



class Backscat(object):
    """ The kinetics of all backscattered ions (energy, location and trajectory)
    """
    pass


class Transmit(object):
    """ The kinetics of all transmitted ions (energy, location and trajectory)
    """
    pass


class Sputter(object):
    """ The kinetics of all target atoms sputtered from the target.
    """
    pass


class Collision:
    """Reads the SRIM Collisions.txt file

    """
    def __init__(self, filename):
        self.filename = filename

        with open(filename, "r", encoding="latin-1") as f:
            self._read_header(f)

        self._ion_index = buffered_findall(filename, b"  Ion    Energy")

    def _read_header(self, f):
        """Read Header of COLLISIONS.TXT

        Currently we do nothing with the header

        """

        # Collect the header of the file
        header = []

        for line in f:
            if line == " \n":
                break
            header.append(line)
        return header

    def _read_ion(self, ion_str):
        """There are 2 types of files with and without cascades

        format:
           1 - Kinchin-Pease Theory (No full cascades)
           2 - full cascades
        """
        # Notice that lines is an generator!
        # This makes it so we can walk through lines
        # in multiple for loops
        lines = (line for line in ion_str.split('\n'))

        # Skip Ion Header
        for line in lines:
            if re.match("^-+\r$", line):
                break

        collisions = []

        # Reads collisions for an ion
        for line in lines:
            if re.match("^=+\r$", line):
                break

            tokens = line.split(chr(179))[1:-1]

            # Check if a full_cascades simulation
            # Read Cascade information
            if re.match(r"\s+<== Start of New Cascade\s+", tokens[-1]):
                (target_disp,
                 target_vac,
                 target_replac,
                 target_inter,
                 cascade) = self._read_cascade(lines)
            else:
                target_disp = float(tokens[8])
                target_vac = 0
                target_replac = 0
                target_inter = 0
                cascade = None

            collisions.append({
                'ion_number': int(tokens[0]),
                'kinetic_energy': float(tokens[1]),
                'depth': float(tokens[2]),
                'lat_y_dist': float(tokens[3]),
                'lat_z_dist': float(tokens[4]),
                'stopping_energy': float(tokens[5]),
                'atom': re.search("([A-Z][a-z]?)", tokens[6]).group(1),
                'recoil_energy': float(tokens[7]),
                'target_disp': target_disp,
                'target_vac': target_vac,
                'target_replac': target_replac,
                'target_inter': target_inter,
                'cascade': cascade
            })

            # Handles weird case where no summary of cascade
            if target_disp is None:
                break;

        # Reads ion footer
        ion_number = re.search(int_regex, next(lines)).group(0)

        footer = ""
        for line in lines:
            if re.match("^=+\r$", line):
                break
            footer += line

        matches = re.findall(double_regex, footer)

        line = next(lines)

        return {
            'ion_number': int(ion_number),
            'displacements': float(matches[0]),
            'avg_displacements': float(matches[1]),
            'replacements': float(matches[2]),
            'avg_replacements': float(matches[3]),
            'vacancies': float(matches[4]),
            'avg_vacancies': float(matches[5]),
            'interstitials': float(matches[6]),
            'avg_interstitials': float(matches[7]),
            'sputtered_atoms': float(matches[8]),
            'avg_sputtered_atoms': float(matches[9]),
            'transmitted_atoms': float(matches[10]),
            'avg_transmitted_atoms': float(matches[11]),
            'collisions': collisions
        }

    def _read_cascade(self, lines):
        line = next(lines)

        assert re.match("^=+\r$", line)


        line = next(lines)
        assert re.match((
                "  Recoil Atom Energy\(eV\)   X \(A\)      Y \(A\)      Z \(A\)"
                "   Vac Repl Ion Numb \d+="
        ), line)

        cascade = []
        for line in lines:
            if re.match("^=+\r$", line):
                break
            tokens = line.split()[1:-1]

            print(tokens)
            cascade.append({
                'recoil': int(tokens[0]),
                'atom': int(tokens[1]),
                'recoil_energy': float(tokens[2]),
                'position': np.array([float(tokens[3]),
                                      float(tokens[4]),
                                      float(tokens[5])]),
                'vac': int(tokens[6]),
                'repl': int(tokens[7])
            })

        if line.count('=') > 100:
            return None, None, None, None, cascade

        line = next(lines)
        tokens = line.split(chr(179))[1:-1]

        if tokens:

            target_disp = float(tokens[2])
            target_vac = float(tokens[3])
            target_replac = float(tokens[4])
            target_inter = float(tokens[5])
        else:
            target_disp = None
            target_vac = None
            target_replac = None
            target_inter = None

        return target_disp, target_vac, target_replac, target_inter, cascade

    def __getitem__(self, i):
        start = self._ion_index[i]

        if i == len(self._ion_index):
            end = os.path.getsize(self.filename)
        else:
            end = self._ion_index[i+1]

        with open(self.filename, "rb") as f:
            f.seek(start)
            # We assume that ion_str will fit in RAM
            ion_str = f.read(end - start)
            return self._read_ion(ion_str.decode('latin-1'))

    def __len__(self):
        return len(self._ion_index) - 1


def buffered_findall(filename, string, start=0):
    with open(filename, 'rb') as f:
        filesize = os.path.getsize(filename)
        BUFFERSIZE = 4096
        overlap = len(string) - 1
        buffer = None
        positions = []

        if start > 0:
            f.seek(start)

        while True:
            if (f.tell() >= overlap and f.tell() < filesize):
                f.seek(f.tell() - overlap)
            buffer = f.read(BUFFERSIZE)
            if buffer:
                buffer_positions = [m.start() for m in re.finditer(string, buffer)]

                for position in buffer_positions:
                    if position >= 0:
                        positions.append(f.tell() - len(buffer) + position)
            else:
                return positions



import yaml
import os
import re

import srim


def create_elementdb():
    dbpath = os.path.join(srim.__path__[0], 'data', 'elements.yaml')
    return yaml.load(open(dbpath, "r"))


class ElementDB(object):
    _db = create_elementdb()

    @classmethod
    def lookup(cls, identifier):
        """ Looks up element from symbol, name, or atomic number

        :param str or int identifier: Unique symbol, name, or atomic number of element
        """
        if isinstance(identifier, (bytes, str)):
            if re.match("^[A-Z][a-z]?$", identifier):   # Symbol
                return cls._lookup_symbol(identifier)
            elif re.match("^[A-Z][a-z]*$", identifier): # Name
                return cls._lookup_name(identifier)
        elif isinstance(identifier, int):               # Atomic Number
            return cls._lookup_atomic_number(identifier)
        raise ValueError('identifier of type:{} value:{} not value see doc'.format(
            type(identifier), identifier))

    @classmethod
    def _lookup_symbol(cls, symbol):
        """ Looks up symbol in element database 

        :param str symbol: Symbol of atomic element
        """
        return cls._db[symbol]

    @classmethod
    def _lookup_name(cls, name):
        """ Looks element in database by name

        :param str name: (Full) Name of atomic element (British spelling)
        """
        for symbol in cls._db:
            if cls._db[symbol]['name'] == name:
                return cls._db[symbol]
        raise KeyError('name:{} does not exist'.format(name))

    @classmethod
    def _lookup_atomic_number(cls, atomic_number):
        """ Look up element in database by atomic number (Z)

        :param int atomic_number: Atomic number of atomic element
        """
        for symbol in cls._db:
            if cls._db[symbol]['z'] == atomic_number:
                return cls._db[symbol]
        raise IndexError('atomic number:{} does not exist'.format(atomic_number))



class Target(object):
    """ Target that Ion impacts

    """
    def __init__(self, layers):
        self.layers = layers

    @property
    def width(self):
        """ total width of target """
        return sum(layer.width for layer in self.layers)



# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='pysrim',
    version='0.2.2',
    description='Srim Automation of Tasks via Python',
    long_description='Pythonic Wrapper to SRIM',
    url='https://gitlab.com/costrouc/srim-python',
    author='Christopher Ostrouchov',
    author_email='chris.ostrouchov+pysrim@gmail.com',
    license="MIT",
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='material srim automation plotting',
    download_url='https://gitlab.com/costrouc/srim-python/repository/master/archive.zip',
    packages=find_packages(exclude=['examples', 'tests', 'test_files', 'docs']),
    package_data={
        'srim': ['data/*.yaml'],
    },
    setup_requires=['pytest-runner', 'setuptools>=38.6.0'],  # >38.6.0 needed for markdown README.md
    install_requires=['pyyaml', 'numpy>=1.10.0'],
    tests_require=['pytest', 'pytest-mock', 'pytest-cov'],
)



import re

from .utils import (
    check_input, 
    is_positive, is_greater_than_zero,
    is_zero_or_one
)
from .element import Element

class Material(object):
    """ Material Representation """
    def __init__(self, elements, density, phase=0):
        """Create Material from elements, density, and phase

        :param dict elements: Dictionary of elements with properties (stoich, E_d, lattice, surface)
        :param float density: Density [g/cm^3] of material
        :params int phase: Phase of material (solid = 0, gas = 1)

        Structure of dictionary elements properties:
        - stoich  (required): Stoichiometry of element (fraction)
        - E_d     (optional): Displacement energy [eV] default 25.0 eV
        - lattice (optional): Lattice binding energies [eV] default 0.0 eV
        - surface (optional): Surface binding energies [eV] default 3.0 eV

        dictionary element properties can be:
        
        float or int: stoich
          all others take default values for now

        dictionary:
          {'stoich', 'E_d', 'lattice', 'surface'}
          stoich is required all others are optional

        elements list structure:
          [stoich, E_d, lattice, surface]
          first element is required all others optional
          
        For example a single element in elements can be specified as:
          - {'Cu': 1.0}
          - {Element('Cu'): 1.0}
          - {Element('Cu'): [1.0, 25.0]}
          - {'Cu': {'stoich': 1.0}}
          - {Element('Cu'): {'stoich': 1.0, 'E_d': 25.0, 'lattice': 0.0, 'surface': 3.0}

        All stoichiometries will be normalized to 1.0

        Eventually the materials will have better defaults that come from databases.
        """
        self.phase = phase
        self.density = density
        self.elements = {}

        stoich_sum = 0.0
        for element in elements:
            values = elements[element]

            if isinstance(values, dict):
                stoich = values['stoich']
                e_disp = values.get('E_d', 25.0)
                lattice = values.get('lattice', 0.0)
                surface = values.get('surface', 3.0)
            elif isinstance(values, list):
                default_values = [0.0, 25.0, 0.0, 3.0]
                if len(values) == 0 or len(values) > 4:
                    raise ValueError('list must be 0 < length < 5')
                values = values + default_values[len(values):]
                stoich, e_disp, lattice, surface = values
            elif isinstance(values, (int, float)):
                stoich = values
                e_disp = 25.0
                lattice = 0.0
                surface = 3.0
            else:
                raise ValueError('elements must be of type int, float, list, or dict')

            # Check input
            stoich = check_input(float, is_greater_than_zero, stoich)
            e_disp = check_input(float, is_positive, e_disp)
            lattice = check_input(float, is_positive, lattice)
            surface = check_input(float, is_positive, surface)
            
            stoich_sum += stoich

            if not isinstance(element, Element):
                element = Element(element)
            
            self.elements.update({element: {
                'stoich': stoich, 'E_d': e_disp,
                'lattice': lattice, 'surface': surface
            }})

        # Normalize the Chemical Composisiton to 1.0
        for element in self.elements:
            self.elements[element]['stoich'] /= stoich_sum


    @classmethod
    def from_formula(cls, chemical_formula, density, phase=0):
        """ Creation Material from chemical formula string and density
        :params str chemical_formula: Chemical formula string in specific format
        :params float density: Density [g/cm^3] of material
        :params int phase: Phase of material (solid = 0, gas = 1)
        Examples of chemical_formula:
         - SiC
         - CO2
         - AuFe1.5
         - Al10.0Fe90.0

        Chemical Formula will be normalized to 1.0. E_d, 
        """
        elements = cls._formula_to_elements(chemical_formula)
        return Material(elements, density, phase)

    @staticmethod
    def _formula_to_elements(chemical_formula):
        """ Convert chemical formula to elements """
        single_element = '([A-Z][a-z]?)([0-9]*(?:\.[0-9]*)?)?'
        elements = {}
        
        if re.match('^(?:{})+$'.format(single_element), chemical_formula):
            matches = re.findall(single_element, chemical_formula)
        else:
            error_str = 'chemical formula string {} does not match regex'
            raise ValueError(error_str.format(chemical_formula))
        
        # Check for errors in stoichiometry
        for symbol, fraction in matches:
            element = Element(symbol)
        
            if element in elements:
                error_str = 'cannot have duplicate elements {} in stoichiometry'
                raise ValueError(error_str.format(element.symbol))
            
            if fraction == '':
                fraction = 1.0
            
            elements.update({element: float(fraction)})
        return elements
        
    @property
    def density(self):
        return self._density

    @density.setter
    def density(self, value):
        self._density = check_input(float, is_positive, value)

    @property
    def phase(self):
        return self._phase

    @phase.setter
    def phase(self, value):
        self._phase = check_input(int, is_zero_or_one, value)
            
    @property
    def chemical_formula(self):
        return ' '.join('{} {:1.2f}'.format(element.symbol, self.elements[element]) for element in self.elements)

    def __repr__(self):
        material_str = "<Material formula:{} density:{:2.3f}>"
        return material_str.format(self.chemical_formula, self.density)

    def __eq__(self, material):
        if abs(self.density - material.density) > 1e-6:
            return False
        
        if len(self.elements) != len(material.elements):
            return False

        for element in self.elements:
            if not element in material.elements:
                return False
            for prop in self.elements[element]:
                if abs(self.elements[element][prop] - material.elements[element][prop]) > 1e-6:
                    return False
        return True



""" Write Inputfile for SRIM calculations

"""


class AutoTRIM(object):
    """ Determines state of TRIM calculation

     - [0] TRIM runs normally.
     - [1] TRIM runs without keyboard input
     - [2] TRIM resumes running its last saved calculation

    [1] is really the only one you would want for python calculations
    """
    def __init__(self, mode=1, restart_directroy=None):
        self._mode = mode

    def write(self):
        """ write AUTOTRIM in current directory """
        with open('TRIMAUTO', 'w') as f:
            f.write('{}'.format(self._mode))


class TRIMInput(object):
    """ Input File representation of TRIM run """
    newline = '\r\n'

    def __init__(self, srim):
        self._srim = srim

    @property
    def srim_num_elements(self):
        return sum(len(layer.elements) for layer in self._srim.target.layers)

    def _write_title(self):
        return (
            'This file controls TRIM Calculations '
            'generated by srim-python'
        ) + self.newline

    def _write_ion(self):
        return (
            'Ion: Z, Mass [amu], Energy [keV], Angle [degrees], '
            'Number Ions, Bragg Corr, AutoSave Number'
        ) + self.newline + '{} {} {} {} {} {} {}'.format(
            self._srim.ion.atomic_number,
            self._srim.ion.mass,
            self._srim.ion.energy / 1000.0, # eV -> keV
            self._srim.settings.angle_ions,
            self._srim.number_ions,
            self._srim.settings.bragg_correction,
            self._srim.settings.autosave
        ) + self.newline

    def _write_cascade_options(self):
        return (
            'Cascades(1=Kitchn-Peese, 2=Full-Cascade, 3=Sputtering, '
            '4-5=Ions;6-7=Neutrons), Random Number Seed, Reminders'
        ) + self.newline + '{} {} {}'.format(
        self._srim.calculation,
            self._srim.settings.random_seed,
            self._srim.settings.reminders
        ) + self.newline

    def _write_plot_on_off(self):
        return (
            'Diskfiles (0=no,1=yes): RANGES.txt, BACKSCATT.txt, '
            'TRANSMIT.txt, Sputtered, COLLISIONS.txt(0=no, 1=Ion, '
            '2=Ion+Recoils), Special EXYZ.txt file'
        ) + self.newline + '{} {} {} {} {} {}'.format(
            self._srim.settings.ranges,
            self._srim.settings.backscattered,
            self._srim.settings.transmit,
            self._srim.settings.sputtered,
            self._srim.settings.collisions,
            self._srim.settings.exyz
        ) + self.newline

    def _write_target(self):
        return (
            'Target material : Number of Elements, Number of Layers'
        ) + self.newline + '"{}" {} {}'.format(
            self._srim.settings.description,
            self.srim_num_elements,
            len(self._srim.target.layers),
        ) + self.newline

    def _write_plot_options(self):
        return (
            'PlotType (0-5); Plot Depths: Xmin, Xmax(Ang.) '
            '[=0 0 for Viewing Full Target]'
        ) + self.newline + '{} {} {}'.format(
            self._srim.settings.plot_mode,
            self._srim.settings.plot_xmin,
            self._srim.settings.plot_xmax
        ) + self.newline

    def _write_elements(self):
        elements_str = (
            'Target Elements:    Z   Mass [amu]'
        ) + self.newline
        index = 1
        for layer in self._srim.target.layers:
            for element in layer.elements:
                elements_str += 'Atom {} = {} =     {} {}'.format(
                    index,
                    element.symbol,
                    element.atomic_number,
                    element.mass
                ) + self.newline
                index += 1
        return elements_str

    def _write_layer(self):
        layer_str = (
            'Layer    Layer Name   Width Density'
        )

        for layer in self._srim.target.layers:
            for element in layer.elements:
                layer_str += '  {}({})'.format(element.symbol, element.atomic_number)

        layer_str += self.newline + (
            'Number   Description  (Ang) (g/cm^3)'
        ) + '  Stoich' * self.srim_num_elements + self.newline


        element_index = 0
        for layer_index, layer in enumerate(self._srim.target.layers, start=1):
            layer_str += '{} "{}" {} {}'.format(
                layer_index,
                layer.name,
                layer.width,
                layer.density
            )
            layer_str += ' 0.0' * element_index
            for element in layer.elements:
                layer_str += ' {} '.format(layer.elements[element]['stoich'])
                layer_str += ' 0.0' * (self.srim_num_elements - element_index - len(layer.elements))
                element_index += len(layer.elements)
        return layer_str + self.newline

    def _write_solid_gas(self):
        return (
            '0  Target layer phases (0=Solid, 1=Gas)'
        ) + self.newline + ' '.join(str(layer.phase) for layer in self._srim.target.layers) + self.newline

    def _write_bragg_correction(self):
        return (
            'Target Compound Corrections (Bragg)'
        ) + self.newline + ' 1' * len(self._srim.target.layers) + self.newline

    def _write_displacement_energies(self):
        ed_str = (
            'Individual target atom displacement energies (eV)'
        ) + self.newline

        for layer in self._srim.target.layers:
            for element in layer.elements:
                ed_str += ' {}'.format(layer.elements[element]['E_d']) + self.newline
        return ed_str

    def _write_lattice_binding(self):
        lattice_str = (
            'Individual target atom lattice binding energies (eV)'
        ) + self.newline

        for layer in self._srim.target.layers:
            for element in layer.elements:
                lattice_str += ' {}'.format(layer.elements[element]['lattice']) + self.newline
        return lattice_str

    def _write_surface_binding(self):
        surface_str = (
            'Individual target atom surface binding energies (eV)'
        ) + self.newline

        for layer in self._srim.target.layers:
            for element in layer.elements:
                surface_str += ' {}'.format(layer.elements[element]['surface']) + self.newline
        return surface_str

    def _write_version(self):
        return (
            'Stopping Power Version (1=2011, 0=2011)'
        ) + self.newline + '{}'.format(self._srim.settings.version) + self.newline

    def write(self):
        with open('TRIM.IN', 'wb') as f:
            methods = [
                self._write_title,
                self._write_ion,
                self._write_cascade_options,
                self._write_plot_on_off,
                self._write_target,
                self._write_plot_options,
                self._write_elements,
                self._write_layer,
                self._write_solid_gas,
                self._write_bragg_correction,
                self._write_displacement_energies,
                self._write_lattice_binding,
                self._write_surface_binding,
                self._write_version
            ]

            input_str = ''
            for method in methods:
                input_str += method.__call__()

            f.write(input_str.encode('utf-8'))


class SRInput(object):
    """ Input file for Stopping and Range (Calculations) """

    newline = '\r\n'

    def __init__(self, sr):
        self._sr = sr

    def _write_filename(self):
        return (
            '---Stopping/Range Input Data (Number-format: Period = Decimal Point)'
        ) + self.newline + (
            '---Output File Name'
        ) + self.newline + '{}'.format(
            self._sr.settings.output_filename
        ) + self.newline

    def _write_ion(self):
        return (
            '---Ion(Z), Ion Mass(u)'
        ) + self.newline + '{} {}'.format(
            self._sr.ion.atomic_number,
            self._sr.ion.mass
        ) + self.newline

    def _write_layer_info(self):
        return (
            '---Target Data: (Solid=0,Gas=1), Density(g/cm3), Compound Corr.'
        ) + self.newline + '{} {} {}'.format(
            self._sr.layer.phase,
            self._sr.layer.density,
            self._sr.settings.correction
        ) + self.newline + (
            '---Number of Target Elements'
        ) + self.newline + '{}'.format(
            len(self._sr.layer.elements)
        ) + self.newline

    def _write_elements(self):
        elements_str = (
            '---Target Elements: (Z), Target name, Stoich, Target Mass(u)'
        ) + self.newline

        for element in self._sr.layer.elements:
            elements_str += '{} "{}" {} {}'.format(
                element.atomic_number,
                element.name,
                self._sr.layer.elements[element]['stoich'],
                element.mass
            ) + self.newline
        return elements_str

    def _write_output_options(self):
        return (
            '---Output Stopping Units (1-8)'
        ) + self.newline + '{}'.format(
            self._sr.settings.output_type
        ) + self.newline

    def _write_ion_energy_range(self):
        return (
            '---Ion Energy : E-Min(keV), E-Max(keV)'
        ) + self.newline + '{} {}'.format(
            self._sr.settings.energy_min / 1.0e3,
            self._sr.ion.energy / 1.0e3
        ) + self.newline

    def write(self):
        with open('SR.IN', 'wb') as f:
            methods = [
                self._write_filename,
                self._write_ion,
                self._write_layer_info,
                self._write_elements,
                self._write_output_options,
                self._write_ion_energy_range
            ]

            input_str = ''
            for method in methods:
                input_str += method.__call__()

            f.write(input_str.encode('utf-8'))



""" Utility functions

"""

def check_input(input_type, condition, value):
    value = input_type(value)
    if not condition(value):
        raise ValueError('type of argument does not satisfy condition')
    return value

is_zero = lambda value: True if value == 0 else False
is_zero_or_one = lambda value: True if value in range(2) else False
is_zero_to_two = lambda value: True if value in range(3) else False
is_zero_to_five = lambda value: True if value in range(6) else False
is_one_to_seven = lambda value: True if value in range(1,8) else False
is_one_to_eight = lambda value: True if value in range(1,9) else False
is_srim_degrees = lambda value: True if 0.0 <= value < 90.0 else False 
is_positive = lambda value: True if value >= 0.0 else False
is_greater_than_zero = lambda value: True if value > 0.0 else False
is_quoteless = lambda value: True if '"' not in value else False
